/*
 * ANTLR Code Generator
 *
 * Take the list of rules and generator code for each alt of every rule.
 *
 * Notes:
 *		ALX has been replaced by DLG (written by Will Cohen).  No names have
 *		been changed in ANTLR, however.
 */

#include <stdio.h>
#ifdef MEMCHK
#include "trax.h"
#endif
#include "set.h"
#include "struct.h"
#ifdef MPW
/*#pragma segment gen*/
#endif

#define MaxLabel		10
#define MaxFileName		30
#define TokFile			"tokens.h"
#define ErrFile			"err.c"
#define LexErrorString "Invalid token"
#define TokenOffset		1	/* Pass 0th position */

#define tabF(f, ta)	{register int i; for (i=1; i<=ta; i++) putc('\t', f);}
#define Tab			tabF(output, tabs)
#define PastWhiteSpace(s)	\
			while (*(s) == ' ' || *(s) == '\t' || *(s) == '\n') {s++;}

static FILE *output = stdout,
			*errtab = NULL;
static int	file=0;
static int	tabs, numErrSets=0;

#define NewLabel	sprintf(label, "_L%d", labelNum++);
static char	label[MaxLabel+1];
static int	labelNum=0;
static FILE *alx;	/* parser.alx output file */

#define NumAlts(b, m)	{register Atom **t=(b)+1; m=0; while ( *t++ != NULL ) {m++;}}
#define MaxDegInFset(f, max)							\
			{	register set *t=(f); unsigned m;		\
				max = 0;								\
				while ( t->n != 0 ) {					\
					m = set_deg(*t);					\
					if ( m > max ) max = m;				\
					t++;								\
				}										\
			}
#define TotalInFset(f, tot)								\
			{	register set *t=(f);					\
				tot = 0;								\
				while ( t->n != 0 ) {					\
					tot += set_deg(*t);					\
					t++;								\
				}										\
			}

int THRif = 3, THRexpr = 3;		/* Alt and First_Set thresholds */

char *StdErrMsg(), *outname(), *StripQuotes();

static
Gen(s,a1,a2,a3)
char *s;
long a1,a2,a3;
{
	Tab;
	fprintf(output, s, a1, a2, a3);
}

static
GenNT(s,a1,a2,a3)
char *s;
long a1,a2,a3;
{
	fprintf(output, s, a1, a2, a3);
}

GenRules(rules)
Atom ***rules;
{
	int r;
	set *f;
	
	errtab = fopen(ErrFile, "w");
	if ( errtab == NULL ) fatal("Can't open errtab file");
	GenErrHdr();
	file = -1;
	output = NULL;
	for (r=1; rules[r] != NULL; r++)
	{
		if ( RuleFile[r] != file )
		{
			if ( output != NULL ) fclose( output );
			output = fopen(outname(FileStr[RuleFile[r]]), "w");
			if ( output == NULL )
			{
				fatal("Can't open output file");
			}
			file = RuleFile[r];
			GenFileHdr();
			if ( r == 1 )
			{
				dmpActList(ActStr);
			}
		}
		f = First(r);
		ambigRule(f, r);
		GRule(rules[r], f, r);
		freeFset(f);
	}
	dmpActList(ActStr2);
	fclose( output );
	if ( numErrSets > 0 ) putc(',', errtab);	/* Put at least this 1 set in table */
	fprintf(errtab, "\n{ 0x%02x }", 0);
	fprintf(errtab, "\n};\n");
	fclose( errtab );
}

GRule(rule, f, r)
Atom **rule;
set *f;
int r;
{
	Atom blk;
	
	/* Create a dummy atom that happens to be a blk atom */
	blk.ptr.block=rule;
	blk.type = aRule;
	blk.parent=NULL;
	blk.inheritance=NULL;
	blk.file = FileStr[RuleFile[r]];
	blk.line = RuleLine[r];
	tabs=0;
	Gen("%s()\n", RuleStr[r]);
	GenBlk(&blk, f, r);
	Gen("\n");
}

GenBlk(atm, f, r)
Atom *atm;
set *f;
int r;
{
	Atom **blk = atm->ptr.block;
	set used;

	set_new(used, MaxToken);
	Gen("{\n");
	if ( atm->type == aRule )
	{
		tabs++;
		Gen("RULEvars;\n");
	}
	Gen("BLKvars;\n");
	if ( blk[1][1].type == aAction )
	{
		dumpAction( blk[1][1].ptr.action, output, tabs );
	}
	Gen("EnterBLK;\n");
	if ( atm->type == aRule ) Gen("EnterRule;\n");
	if ( MakeTrax && atm->type == aRule ) Gen("aTrax(\"%s\",LexText);\n", RuleStr[r]);
	switch ( atm->type )
	{
		case aRule :
		case aSubBLK :	GenSubBLK(atm, f, r, &used); break;
		case aOptBLK :	GenOptBLK(atm, f, r, &used); break;
		case aLoopBLK :	GenLoopBLK(atm, f, r, &used); break;
	}
	if ( atm->type == aRule )
	{
		Gen("SUCCESS;\n");
		--tabs;
		Gen("Fail: ;\n");
		tabs++;
		if ( FailActionStr[r] != NULL )
		{
			dumpAction( FailActionStr[r], output, tabs );
		}
		Gen("FAILURE;\n");
		--tabs;
	}
	else Gen("ExitBLK;\n");
	Gen("}\n");
	set_free( used );
}

/*
 * Some rules for generating optimized IF sequences:
 *
 *	SubBLK:
 *		if ( #alts == 1 ) don't generate "if (First(alt))"
 *		if ( #alts == 1 && 1st element==aLeaf ) use "if (!lex(token))" not Munch
 *	OptBLK:
 *		Always generate "if (First(alt))"
 *		Always gen Munch if 1st element==aLeaf
 *	LoopBLK:
 *		if ( #alts == 1 ) don't generate "if (First(alt))"
 *		Always gen Munch if 1st element==aLeaf
 */
GenIF(atm, f, r, used, m)
Atom *atm;
set *f, *used;
int r,m;	/* m is # of alts in blk */
{
	register Atom **blk = atm->ptr.block;
	int alt, at, start;

	for (alt=1; blk[alt] != NULL; alt++)
	{
		if ( atm->type == aOptBLK || 
			 ((atm->type==aLoopBLK||atm->type==aSubBLK||atm->type==aRule) && m > 1) )
		{
			if ( alt > 1 ) Gen("else if ( ");
			else Gen("if ( ");
			GenIFexpr( f[alt-1], used, r, alt, atm );
			GenNT(" ) {\n");
			tabs++;
		}
		start=1;
		if ( alt==1 && blk[1][1].type == aAction ) start=2;
		for (at=start; blk[alt][at].type != EOL; at++)
		{
			if ( at==start && blk[alt][1].type == aLeaf )
			{
				if ( atm->type == aOptBLK || atm->type == aLoopBLK ||
					 ((atm->type==aRule||atm->type==aSubBLK) && m>1) )
				{
					Gen("Munch;\n");
					continue;
				}
			}
			GenAtom( &(blk[alt][at]), r );
		}
		if ( atm->type == aOptBLK || 
			 ((atm->type==aLoopBLK||atm->type==aSubBLK||atm->type==aRule) && m > 1) )
		{
			--tabs;
			Gen("}\n");
		}
	}
}

GenIFexpr(f, used, r, m, atm)
set f, *used;
Atom *atm;
int r, m;
{
	unsigned *e, *g, firstToken=1, n=0;

	if ( set_deg(f) == 0 )
	{
		if ( m != -1 )
		{
			if ( atm->type == aRule )
			{
				warnFL( eMsg(
					"alt %d rule %s empty first set",
					 m, RuleStr[r]),
					 FileStr[RuleFile[r]], RuleLine[r] );
			}
			else
			{
				errBlk( (*atm), eMsg( " [rule %s] alt %d empty first set",
							 RuleStr[r], m) );
			}
		}
		else
		{
			errBlk( (*atm), eMsg( " [rule %s] optional path has empty first set",
							 RuleStr[r]) );
		}
		return;
	}
	if ( (e=g=set_pdq(f)) == NULL )
	{
		fatal("Cannot allocate IF expr pdq set");
	}
	while ( *e != nil )
	{
		if ( !set_el(*e, *used) )	/* Resolve ambig by using token only 1st time */
		{
			if ( !firstToken ) GenNT(" || ");
			else firstToken = 0;
			GenNT("Token==%s", TokenValue(*e));
			set_orel(*e, used);
			n++;
		}
		e++;
	}
	free( g );
	if ( n == 0 )
	{
		GenNT("0");
		if ( m != -1 )
		{
			if ( atm->type == aRule )
			{
				warnFL( eMsg(
					"warning: alt %d rule %s unreachable",
					 m, RuleStr[r]),
					 FileStr[RuleFile[r]], RuleLine[r] );
			}
			else
			{
				warnBlk( (*atm), eMsg(
					" [rule %s] alt %d unreachable",
					 RuleStr[r], m) );
			}
		}
		else
		{
			warnBlk( (*atm), eMsg(
			 	" [rule %s] optional path disappears",
				 RuleStr[r]) );
		}
	}
}

GenSubBLK(atm, f, r, used)
Atom *atm;
set *f;
int r;
set *used;
{
	register Atom **blk = atm->ptr.block;
	int	m,n;
	
	NumAlts(blk, m);
	MaxDegInFset(f, n);
	/* Don't gen outsidef first set membership code if only one alternative
	 * 'cause no matter what we should enter that single alternative.
	 */
	if ( m <= THRif && n <= THRexpr || m == 1 )	/* Optimize, we are below threshold */
	{
		GenIF(atm, f, r, used, m);
		if ( m > 1 ) Gen("else {%s goto Fail;}\n", StdErrMsg(f, FALSE));
	}
	else
	{
		GenSw(atm, f, r, used);
		GenSwClose( f );
	}
}

GenOptBLK(atm, f, r, used)
Atom *atm;
set *f;
int r;
set *used;
{
	register Atom **blk = atm->ptr.block;
	int	m,n;
	
	NumAlts(blk, m);
	MaxDegInFset(f, n);
	if ( m <= THRif && n <= THRexpr )	/* Optimize, we are below threshold */
	{
		GenIF(atm, f, r, used, m);
		Gen("else if (!( ");
		GenIFexpr( f[m], used, r, -1, atm );
		GenNT(" )) {%s goto Fail;}\n", StdErrMsg(f, FALSE));
	}
	else
	{
		GenSw(atm, f, r, used);
		GenCase( f[m], used, r, -1, atm );
		tabs++;
		Gen("break;\n");
		--tabs;
		GenSwClose( f );
	}
}

GenLoopBLK(atm, f, r, used)
Atom *atm;
set *f;
int r;
set *used;
{
	register Atom **blk = atm->ptr.block;
	int	m,n;
	
	NumAlts(blk, m);
	MaxDegInFset(f, n);
	/* If only one alt and expr is simple enough, use "while (First(alt))" */
	if ( m == 1 && n <= THRexpr )
	{
		Gen("while ( ");
		GenIFexpr( f[0], used, r, 1, atm );
		GenNT(" )\n");
		Gen("{\n");
		tabs++;
		GenIF(atm, f, r, used, m);
		Gen("sREL;\n");
		tabs--;
		Gen("}\n");
		Gen("if (!( ");
		GenIFexpr( f[m], used, r, -1, atm );
		GenNT(" )) {%s goto Fail;}\n", StdErrMsg(f, FALSE));
		return;
	}
	Gen("do {\n");
	tabs++;
	if ( m <= THRif && n <= THRexpr )	/* Optimize, we are below threshold */
	{
		GenIF(atm, f, r, used, m);
		NewLabel;
		Gen("else goto %s;\n", label);
	}
	else
	{
		GenSw(atm, f, r, used);
		GenCase( f[m], used, r, -1, atm );
		NewLabel;
		tabs++;
		Gen("goto %s;\n", label);
		--tabs;
		GenSwClose( f );
	}
	Gen("sREL;\n");
	--tabs;
	Gen("} while (1);\n");
	Gen("%s: ;\n", label);
	if ( m <= THRif && n <= THRexpr )
	{
		Gen("if (!( ");
		GenIFexpr( f[m], used, r, -1, atm );
		GenNT(" )) {%s goto Fail;}\n", StdErrMsg(f, FALSE));
	}
}

GenSw(atm, f, r, used)
Atom *atm;
set *f, *used;
int r;
{
	Atom **blk = atm->ptr.block;
	int alt, at, start;
	
	Gen("switch ( Token )\n");
	Gen("{\n");
	for (alt=1; blk[alt] != NULL; alt++)
	{
		GenCase( f[alt-1], used, r, alt, atm );
		tabs++;
		start=1;
		if ( alt==1 && blk[1][1].type == aAction ) start=2;
		for (at=start; blk[alt][at].type != EOL; at++)
		{
			if ( at==start && blk[alt][1].type == aLeaf )
			{
				Gen("Munch;\n");
				continue;
			}
			GenAtom( &(blk[alt][at]), r );
		}
		Gen("break;\n");
		--tabs;
	}
}

GenSwClose( f )
set *f;
{
	Gen("default : %s goto Fail;\n", StdErrMsg(f, FALSE));
	Gen("}\n");
}

GenAtom(a, r)
Atom *a;
int r;
{
	set *f;

	switch ( a->type )
	{
		case aLeaf :
			if ( RegExprStr[a->token] != NULL )
				if ( RegExprStr[a->token][1] == '@' )
				{
					if ( (a+1)->type != EOL && (a+1)->type != aAction )
					{
						warnFL( eMsg("rule %s elements beyond EOF (\"@\") unreachable",
								RuleStr[r]), a->file, a->line );
					}
				}
			Gen("if ( !lex(%s) ) { %s goto Fail; }\n", TokenValue(a->token),
											StdErrMsg(a->token,TRUE));
			break;
		case aRule :
			/* warnRule(a, r); */
			if ( a->inheritance == NULL )
			{
				if ( ForceInherit )
				{
					Gen("aPush(aArg(0));\n");
					Gen("if (!%s()) goto Fail;\n", a->ptr.rule);
				}
				else
				{
					Gen("Make0; if (!%s()) goto Fail;\n", a->ptr.rule);
				}
			}
			else
			{
				GenBlkPreamble(a);
				Gen("if (!%s()) goto Fail;\n", a->ptr.rule);
				if ( a->inheritance[0] == '=' )
				{
					tabs--;
					Gen("}\n");
				}
			}
			break;
		case aAction :
			dumpAction( a->ptr.action, output, tabs );
			break;
		case aSubBLK :
		case aOptBLK :
		case aLoopBLK :
			if ( a->ptr.block == NULL ) break;
			if ( emptyBlk(a->ptr.block) ) break;
            if ( a->inheritance == NULL )
            {
                if ( ForceInherit )
                {
                    Gen("aPush(aArg(0));\n");
                }
                else Gen("Make0;\n");
            }
            else GenBlkPreamble(a);
			fAtom(a,r,FALSE,&f);
			ambigSub(f, a, r);
			GenBlk(a, f, r);
			if ( a->inheritance != NULL )
				if ( a->inheritance[0] == '=' )
				{
					tabs--;
					Gen("}\n");
				}
			freeFset(f);
			break;
	}
}

GenBlkPreamble(a)
Atom *a;
{
	switch ( a->inheritance[0] )
	{
		case '{' :	Gen("Make0;\n");
					Gen("%s\n", a->inheritance);
					break;
		case '=' :	Gen("{\n");
					tabs++;
					Gen("static Attrib inh%s;\n", a->inheritance);
					Gen("aPush(inh);\n");
					break;
		default  :	Gen("aPush(%s);\n", a->inheritance);
					break;
	}
}

GenCase( f, used, r, m, atm )
set f;
set *used;
int r,m;
Atom *atm;
{
	unsigned *e,*g;
	char *s;
	unsigned n=0;

	if ( set_deg(f) == 0 )
	{
		if ( m != -1 )
		{
			if ( atm->type == aRule )
			{
				warnFL( eMsg(
					"alt %d rule %s empty first set",
					 m, RuleStr[r]),
					 FileStr[RuleFile[r]], RuleLine[r] );
			}
			else
			{
				errBlk( (*atm), eMsg( " [rule %s] alt %d empty first set",
							 RuleStr[r], m) );
			}
		}
		else
		{
			errBlk( (*atm), eMsg( " [rule %s] optional path has empty first set",
							 RuleStr[r]) );
		}
		return;
	}
	if ( (e=g=set_pdq(f)) == NULL )
	{
		fatal("Cannot allocate IF expr pdq set");
	}
	while ( *e != nil )
	{
		if ( !set_el(*e, *used) )
		{
			s = TokenValue(*e);
			if ( *s > '9' || *s < '0' ) Gen("case %s :\n", s);
			else Gen("case %s : /* %s */\n", s, TokenName(*e));
			set_orel(*e, used);
			n++;
		}
		e++;
	}
	free( g );
	if ( n == 0 )
	{
		if ( m != -1 )
		{
			warnBlk( (*atm), eMsg(
			 	"[rule %s] alt %d unreachable",
				 RuleStr[r], m) );
		}
		else
		{
			warnBlk( (*atm), eMsg(
			 	"[rule %s] optional path disappears",
				 RuleStr[r]) );
		}
	}
}

GenErrHdr()
{
	char *p;
	int i;

	fprintf(errtab, "/*\n");
	fprintf(errtab, " * A n t l r  E r r o r  T a b l e  H e a d e r\n");
	fprintf(errtab, " *\n");
	fprintf(errtab, " * Generated from:");
	for (i=0; i<NumFiles; i++) fprintf(errtab, " %s", FileStr[i]);
	fprintf(errtab, "\n");
	fprintf(errtab, " *\n");
	fprintf(errtab, " * (c) 1989, 1990 by Terence Parr, Hank Dietz and Will Cohen\n");
	fprintf(errtab, " * Purdue University Electrical Engineering\n");
	fprintf(errtab, " * ANTLR Version 1.0B\n");
	fprintf(errtab, " */\n\n");

	fprintf(errtab, "#include <stdio.h>\n");
	fprintf(errtab, "#include \"dlgdef.h\"\n");
	fprintf(errtab, "#define nil ((unsigned)-1)\n");
	fprintf(errtab, "#ifdef MPW		/* Macintosh Programmer's Workshop */\n");
	fprintf(errtab, "#define ErrHdr \"File \\\"%%s\\\"; Line %%d #\"\n");
	fprintf(errtab, "#else\n");
	fprintf(errtab, "#define ErrHdr \"\\\"%%s\\\", line %%d:\"\n");
	fprintf(errtab, "#endif\n\n");
	fprintf(errtab, "extern char *tokens[];\n");
	fprintf(errtab, "char *aSourceFile = \"stdin\";\t/* Set to current src filename */\n\n");
	fprintf(errtab, "unsigned *err_decode(p)\n");
	fprintf(errtab, "register unsigned char *p;\n");
	fprintf(errtab, "{\n");
	fprintf(errtab, "\tstatic unsigned bitmask[] = {\n");
	fprintf(errtab, "\t	0x00000001, 0x00000002, 0x00000004, 0x00000008,\n");
	fprintf(errtab, "\t	0x00000010, 0x00000020, 0x00000040, 0x00000080,\n");
	fprintf(errtab, "\t};\n");
	fprintf(errtab, "\textern unsigned *malloc();\n");
	fprintf(errtab, "\tregister unsigned char *endp = &(p[%d]);\n",
					NumWords(MaxToken)*BytesPerWord);
	fprintf(errtab, "\tregister unsigned *q,*r,e= 0;\n\n");
	fprintf(errtab, "\tr = q = (unsigned *) malloc(%d*sizeof(unsigned));\n",
					(NumWords(MaxToken)*WORDSIZE)+1);
	fprintf(errtab, "\tif ( q == 0 ) return( 0 );\n");
	fprintf(errtab, "\tdo {\n");
	fprintf(errtab, "\t	register unsigned t = *p;\n");
	fprintf(errtab, "\t	register unsigned *b = &(bitmask[0]);\n");
	fprintf(errtab, "\t	do {\n");
	fprintf(errtab, "\t		if ( t & *b ) *q++ = e;\n");
	fprintf(errtab, "\t		e++;\n");
	fprintf(errtab, "\t	} while (++b < &(bitmask[%d]));\n", 8);	/* 8 bits/byte */
	fprintf(errtab, "\t} while (++p < endp);\n");
	fprintf(errtab, "\t*q = nil;\n");
	fprintf(errtab, "\treturn( r );\n");
	fprintf(errtab, "}\n\n");

	fprintf(errtab, "err_print(e)\n");
	fprintf(errtab, "unsigned char *e;\n");
	fprintf(errtab, "{\n");
	fprintf(errtab, "	unsigned *p, *q;\n\n");
	fprintf(errtab, "	if ( (q=p=err_decode(e))==NULL ) return;\n");
	fprintf(errtab, "	fprintf(stderr, \"{\");\n");
	fprintf(errtab, "	while ( *p != nil )\n");
	fprintf(errtab, "	{\n");
	fprintf(errtab, "		fprintf(stderr, \" %%s\", tokens[*p++]);\n");
	fprintf(errtab, "	}\n");
	fprintf(errtab, "	fprintf(stderr, \" }\");\n");
	fprintf(errtab, "	free(q);\n");
	fprintf(errtab, "}\n\n");

	fprintf(errtab, "dfltError(e, single)\n");
	fprintf(errtab, "union { unsigned char *eset; unsigned tok; } e;\n");
	fprintf(errtab, "int single;\n");
	fprintf(errtab, "{\n");
	fprintf(errtab, "\tfprintf(stderr, ErrHdr, aSourceFile, lex_line);\n");
	fprintf(errtab, "\tif ( single )\n");
	fprintf(errtab, "\t{\n");
	fprintf(errtab, "\t\tfprintf(stderr, \" missing %%s (Was \\\"%%s\\\"(%%s))\\n\",\n");
	fprintf(errtab, "\t\t\t\t\t\ttokens[(int)e.tok],\n");
	fprintf(errtab, "\t\t\t\t\t\t(tokens[Token][0]=='@')?\"eof\":LexText,\n");
	fprintf(errtab, "\t\t\t\t\t\ttokens[Token]);\n");
	fprintf(errtab, "\t\treturn;\n");
	fprintf(errtab, "\t}\n");
	fprintf(errtab, "\tfprintf(stderr, \" missing \");\n");
	fprintf(errtab, "\terr_print(e.eset);\n");
	fprintf(errtab, "\tfprintf(stderr, \" (Was \\\"%%s\\\"(%%s))\\n\",\n\t\t(tokens[Token][0]=='@')?\"eof\":LexText, tokens[Token]);\n");
	fprintf(errtab, "}\n\n");
	
	fprintf(errtab, "char *tokens[%d]={\n", MaxToken);
	fprintf(errtab, "\t/* 00 */\t\"Invalid\"");
	for (i=TokenOffset; i<=MaxToken; i++)
	{
		if ( i == EpToken ) continue;
		p = TokenName(i);
		if ( *p == '"' )
			fprintf(errtab, ",\n\t/* %02d */\t%s", i, p);
		else
			fprintf(errtab, ",\n\t/* %02d */\t\"%s\"", i, p);
	}
	fprintf(errtab, "\n};\n");
	fprintf(errtab, "unsigned char err[][%d]={", NumWords(MaxToken)*BytesPerWord);
}

GenFileHdr()
{
	Gen("/*\n");
	Gen(" * A n t l r  T r a n s l a t i o n  H e a d e r\n");
	Gen(" *\n");
	Gen(" * This file translated from: %s\n", FileStr[file]);
	Gen(" *\n");
	Gen(" * (c) 1989, 1990 by Terence Parr, Hank Dietz and Will Cohen\n");
	Gen(" * Purdue University Electrical Engineering\n");
	Gen(" * ANTLR Version 1.0B\n");
	Gen(" */\n");
	Gen("#include <stdio.h>\n");
    if ( DfltErr )
        Gen("#define aError(a,b)    dfltError(a,b)\n");
	if ( AttribAction == NULL && file == 0 )
	{
		warn("warning: no attribute type has been defined");
	}
	dumpAction( AttribAction, output, tabs );
	Gen("#include \"dlgdef.h\"\n");
	Gen("#include \"attrib.h\"\n");
	Gen("#include \"%s\"\n", TokFile);
	Gen("extern unsigned char err[][%d];\n", NumWords(MaxToken)*BytesPerWord);
	Gen("extern char *tokens[];\n");
	if ( file == 0 )
	{
		Gen("#ifndef aStackSize\n");
		Gen("#define aStackSize 200\n");
		Gen("#endif\n");
		Gen("Attrib\taStack[aStackSize];\n");
		Gen("int\t\t_asp=aStackSize;\n");
	}
	Gen("\n");
}

char *
StdErrMsg( f, single )
union { set *fset; unsigned token; } f;
int single;
{
	static char buf[100];
	unsigned nextErrSet();
	
	if ( single )
	{
		sprintf(buf, "aError(%s,1);", TokenValue(f.token));
	}
	else
	{
		sprintf(buf, "aError(&(err[%d][0]),0);", nextErrSet(f.fset));
	}
	return( &(buf[0]) );
}

static
unsigned
nextErrSet(f)
set *f;
{
	set a;
	register unsigned *p, *endp;

	a = fCompact(f);
	if ( a.n == 0 )
	{
		if ( numErrSets > 0 ) putc(',', errtab);
		fprintf(errtab, "\n{ 0x%08x }", 0);
		return(numErrSets++);
	}
	p = a.setword;
	endp = &(a.setword[a.n]);
	if ( numErrSets > 0 ) putc(',', errtab);
	fprintf(errtab, "\n{ ");
	pErrSet(a);
	fprintf(errtab, " }");

	set_free(a);
	return( numErrSets++ );
}

pErrSet(a)
set a;
{
	register unsigned char *p,*endp,*begin;
	int e=0;

	begin=p=set_bytes(a);
	endp = &(p[a.n*BytesPerWord]);
	while ( p < endp )
	{
		if ( e > 0 ) fprintf(errtab, ", ");
		fprintf(errtab, "0x%02x", (unsigned)*p++);
		e++;
	}
	free(begin);
}

/*
 * Return array of unsigned char which is 8bit-word bitset representation of
 * or WORDSIZEbit-word bitset repr.  We need to convert to bytes, cause we
 * don't know wordsize of target machine, but generally unsigned char is
 * 8 bits.
 *
 * The array returned has BytesPerWord*a.n elements.
 * The ith byte is masked off within each word in the set.  That isolated
 * byte is shifted down into the first 8 bits for placing it into a char.
 */
unsigned char *
set_bytes(a)
set a;
{
	static unsigned mask[4] = {0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000};
	unsigned char *tab, *t;
	register unsigned *p = a.setword;
	register unsigned *endp = &(a.setword[a.n]);
	
	t = tab = (unsigned char *) malloc( a.n*BytesPerWord );
	if ( tab == NULL ) return( NULL );
	do {
		register unsigned *b = &(mask[0]);
		do {
			*t++ = (*p & *b) >> (8*(b-&(mask[0])));
		} while ( ++b < &(mask[BytesPerWord]) );
		p++;
	} while ( p < endp );
	return( tab );
}

dmpActList(actions)
char **actions;
{
	int a;
	
	for (a=1; actions[a] != NULL; a++)
	{
		dumpAction(actions[a], output, 0);
	}
}

static
dumpAction(s, output, tabs)
char *s;
FILE *output;
int tabs;
{
    int inDQuote, inSQuote;
    
    if ( s == NULL ) return;
	Tab;
    PastWhiteSpace( s );
    inDQuote = inSQuote = FALSE;
    while ( *s != '\0' )
    {
        if ( *s == '\\' )
        {
            putc( *s++, output ); /* Avoid '"' Case */
            if ( *s == '\0' ) return;
            if ( *s == '\'' ) putc( *s++, output );
            if ( *s == '\"' ) putc( *s++, output );
        }
        if ( *s == '\'' )
        {
            if ( !inDQuote ) inSQuote = !inSQuote;
        }
        if ( *s == '"' )
        {
            if ( !inSQuote ) inDQuote = !inDQuote;
        }
        if ( *s == '\n' )
        {
            putc('\n', output);
            PastWhiteSpace( s );
            if ( *s == '}' )
            {
                --tabs;
				Tab;
                putc( *s++, output );
                continue;
            }
            if ( *s == '\0' ) return;
			if ( *s != '#' )	/* #define, #endif etc.. start at col 1 */
            {
				Tab;
			}
        }
        if ( *s == '}' && !(inSQuote || inDQuote) )
        {
            --tabs;            /* Indent one fewer */
        }
        if ( *s == '{' && !(inSQuote || inDQuote) )
        {
            tabs++;            /* Indent one more */
        }
        putc( *s, output );
        s++;
    }
    putc('\n', output);
}

/*
 * Return a string corresponding to the output file name associated
 * with the input file name passed in.
 *
 * Observe the following rules:
 *
 *		f.e		--> f".c"
 *		f		--> f".c"
 *		f.		--> f".c"
 *		f.e.g	--> f.e".c"
 *
 * Where f,e,g are arbitrarily long sequences of characters in a file
 * name.
 *
 * In other words, if a ".x" appears on the end of a file name, make it
 * ".c".  If no ".x" appears, append ".c" to the end of the file name.
 */
char *
outname( fs )
char *fs;
{
	static char buf[MaxFileName+1];
	char *p;
	
	p = buf;
	strcpy(buf, fs);
	while ( *p != '\0' )  {p++;}			/* Stop on '\0' */
	while ( *p != '.' && p != buf ) {--p;}	/* Find '.' */
	if ( p != buf ) *p = '\0';				/* Found '.' */
	strcat(buf, ".c");
	return( buf );
}

/* A L X  S t u f f */

InitAlxFile()
{
	if ( LexFile == NULL ) return;
	alx = fopen(LexFile, "w");
	if ( alx == NULL ) fatal( eMsg("Can't open %s\n", LexFile) );
	DmpAlxHdr();
}

CloseAlxFile()
{
	if ( LexFile == NULL ) return;
	fclose( alx );
}

AlxSep()
{
	if ( LexFile == NULL ) return;
	fprintf(alx, "\n%%%%\n\n");
}

DmpAlxAction(a, t)
char *a;
int t;
{
	if ( LexFile == NULL ) return;
	fprintf(alx, "<<\n");
	dumpAction( a, alx, t );
	fprintf(alx, ">>\n");
}

DmpAlxHdr()
{
	fprintf(alx, "<<\n");
	fprintf(alx, "/* %s -- DLG Description of scanner\n", LexFile);
	fprintf(alx, " *\n");
	fprintf(alx, " * Generated from:");
	{int i; for (i=0; i<NumFiles; i++) fprintf(alx, " %s", FileStr[i]);}
	fprintf(alx, "\n");
	fprintf(alx, " *\n");
	fprintf(alx, " * (c) 1989, 1990 by Terence Parr, Hank Dietz and Will Cohen\n");
	fprintf(alx, " * Purdue University Electrical Engineering\n");
	fprintf(alx, " * ANTLR Version 1.0B\n");
	fprintf(alx, " */\n");
	fprintf(alx, "#include <stdio.h>\n");
	fprintf(alx, "#include \"%s\"\n", TokFile);
	fprintf(alx, "int Token;\n");
	dumpAction( AttribAction, alx, 0 );
	fprintf(alx, "#include \"attrib.h\"\n");
	fprintf(alx, "#ifndef LEX_BUF\n");
	fprintf(alx, "#define LEX_BUF 2000\n");
	fprintf(alx, "#endif\n");
	fprintf(alx, "action0()\n");
	fprintf(alx, "{\n");
	fprintf(alx, "\tlex_err(\"%s\");\n", LexErrorString);
	fprintf(alx, "\tadvance();\n");
	fprintf(alx, "\tLexSkip();\n");
	fprintf(alx, "}\n\n");
	fprintf(alx, "lex(t)\n");
	fprintf(alx, "int t;\n");
	fprintf(alx, "{\n\textern char LexText[];\n");
	fprintf(alx, "\tif ( Token == t )\n");
	fprintf(alx, "\t{\n");
	fprintf(alx, "\t\tMakeAttr;\n");
	fprintf(alx, "\t\tif ( Token != Eof ) GetToken();\n");
	fprintf(alx, "\t\treturn( 1 );\n");
	fprintf(alx, "\t}\n");
	fprintf(alx, "\treturn( 0 );\n");
	fprintf(alx, "}\n");
	fprintf(alx, "\>\>\n");
}

DumpTokens( tokens )
char **tokens;
{
	FILE *f;
	int i;

#ifdef OLD
	tokens += TokenOffset;
#endif
	f = fopen(TokFile, "w");
	if ( f == NULL ) fatal("Can't open token file");
	fprintf(f, "/* %s -- List of labelled tokens\n", TokFile);
	fprintf(f, " *\n");
	fprintf(f, " * Generated from:");
	{int i; for (i=0; i<NumFiles; i++) fprintf(f, " %s", FileStr[i]);}
	fprintf(f, "\n");
	fprintf(f, " *\n");
	fprintf(f, " * (c) 1989, 1990 by Terence Parr, Hank Dietz and Will Cohen\n");
	fprintf(f, " * Purdue University Electrical Engineering\n");
	fprintf(f, " * ANTLR Version 1.0B\n");
	fprintf(f, " */\n");
#ifdef OLD
	i = TokenOffset;
	while ( *tokens != NULL )
	{
		if ( **tokens != '"' && i != EpToken )
		{
			fprintf(f, "#define %s %d\n", *tokens, i);
		}
		tokens++;
		i++;
	}
#endif
	for (i=TokenOffset; i<=MaxToken; i++)
	{
		if ( tokens[i] != NULL && i != EpToken )
			fprintf(f, "#define %s %d\n", tokens[i], i);
	}
	fclose( f );
}

DmpAlxRule(rexpr, label, tokenNumber, action)
char *rexpr, *label;
int tokenNumber;
char *action;
{
	if ( LexFile == NULL ) return;
	if ( rexpr == NULL ) fatal("Hmmm...No regular expression to generate");
	if ( *rexpr != '"' ) return;
	fprintf(alx, "%s\n\t<<\n", StripQuotes(rexpr));
	if ( label != NULL )
	{
		fprintf(alx, "\t\tToken = %s;\n", label);
	}
	else
	{
		fprintf(alx, "\t\tToken = %d;\n", tokenNumber);
	}
	if ( action != NULL )
	{
		dumpAction( action, alx, 2 );
	}
	fprintf(alx, "\t>>\n\n");
}

char *
StripQuotes(s)
char *s;
{
	static char buf[MaxTokenSize+1];
	
	if ( strlen(s) >= MaxTokenSize )
	{
		fatal( eMsg("StripQuotes(): Out of buf space (%d char)\n", MaxTokenSize) );
	}
	if ( *s == '"' )
	{
		strcpy(buf, s+1);
		buf[ strlen(buf)-1 ] = '\0';    /* remove last quote */
		return( buf );
	}
	return( s );
}

/*
 * Return the token value of the token sent in
 *
 * If no token label then return number (string)
 * If token name avail then return token name string
 */
char *
TokenValue( t )
int t;
{
	static char numBuf[10];
	
	if ( TokenStr[t] != NULL ) return( TokenStr[t] );
	sprintf(numBuf, "%d", t);
	return( numBuf );
}

char *
TokenName( t )
int t;
{
	if ( TokenStr[t] != NULL ) return( TokenStr[t] );
	if ( RegExprStr[t] != NULL ) return( RegExprStr[t] );
	return("Invalid");
}

ContainsChar(s, c)
register char *s, c;
{
	while ( *s != c && *s != '\0' ) s++;
	if ( *s == c ) return( TRUE );
	return( FALSE );
}
