/* main routine and dfa constructor.
 * ee595e project
 *
 * Will Cohen
 * 10/4/89
 */

#include	<stdio.h>
#include	"dlg.h"
#include	"tokens.h"

set	used_chars;	/* marks characters that are actually used in spec */
set	used_classes;	/* marks characters that are actually used in spec */


path	stack[MAX_STACK];	/* used by parser to build nfa */
int	sp = 0;

int		nfa_pool = 0;	/* last nfa node made */
int		dfa_pool = 0;	/* last dfa node made */
				/* also head of undone queue */
int		last_done = 0;	/* last dfa node examined */

nfa_node	*nfa_array;	/* first nfa node in array */
dfa_node	*dfa_array;	/* first dfa node in array */

nfa_node	clear_nfa_node;	/* prototype of node used for init. */
dfa_node	clear_dfa_node;	/* prototype of node used for init. */

int	nfa_set = 0;	/* used so operation done only once even */
			/* though there might be multiple arc to it */
int	dfa_set = 0;	/* same as nfa_set */

int		nfa_state = 0;	/* used to identify each nfa node */
int		dfa_state = 0;	/* used to identify each dfa node */

int		*dfa_hash_table;

/************ flags and options ***********/
int	compress = FALSE;	/* do character class compression */
int	num_of_nfa = SETSIZE;
int	num_of_dfa = SETSIZE;
int	hash_entries = 2*SETSIZE;

/************ file stuff ******************/
FILE *infile;
FILE *outfile;

char *in_file_name;	/* used to report errors */

main(argc, argv)
int argc;
char **argv;
{
	int c,p;

	init_things();
	/* find out options set */
	options(argc, argv);

	init_arrays();

	/* open input, first thing other after first arg without '-' */
	c = 1; argv++;
	while ((c<argc) && (**argv=='-')){
		c++;
		argv++;
		}
	/* check to make sure valid */
	if ((c>=argc) || (**argv=='-'))
		usage();
	in_file_name = *argv;
	infile = fopen(in_file_name,"r");
	if (!infile)
		printf("dlg: Warning: Can't find file %s.\n",*argv);
	/* open output, second thing other after first arg without '-' */
	c++; argv++;
	while ((c<argc) && (**argv=='-')){
		c++;
		argv++;
		}
	/* if check to make sure valid */
	if ((c>=argc) || (**argv=='-'))
		usage();
	outfile = fopen(*argv,"w");
	if (!outfile)
		printf("dlg: Warning: Can't write to file %s.\n",*argv);
	if ((!infile) || (!outfile))
		exit(2);


	
	/* now time to parse input and build nfa graph */
	advance(); /* prime the input */
	if (grammar())
		warning("Syntax error in input specifications ", line_no);
	/* close any loose ends */
	print_tail();
}

nfa_to_dfa(start)
int start;
{
	set x, t, y, classes;
	unsigned *place;	/* keeps track of where in set */
	int d_state;		/* index to dfa state working on*/
	int trans_d_state;	/* state going to on transistion */
	int a;

	x = closure(set_of(start));
	/* Make x a dfa state */
	d_state = dfastate(x);
	last_done = d_state;
	do {
		/* Mark dfa state x as "done" */
		dfa_array[d_state].done = TRUE;
		last_done ++;	/* move forward in queue */
		/* each "class" actually used in alphabet */
		classes = used_classes;
		place = &(classes.setword[0]);
		set_remove_int(classes,a,place);
		while (a != nil){
			/* Add NFA states reached by a... */
			/* each state in x */
			t = reach(dfa_array[d_state].nfa_states,a);
			/* Were any states found? */
			if (!set_nil(t)) {
				/* Yes, compute closure */
				y = closure(t);
				/* Make DFA state of it ... */
				trans_d_state = dfastate(y);
				/* And make transition x->y, labeled with a */
				dfa_array[d_state].trans[a] =
					(short) trans_d_state;
				}
			set_remove_int(classes,a,place);
			}
		/* now check to see if some dfa nodes still need to be
		   done are in the queue. Repeat until none left */
		d_state = last_done;
		} while (last_done<dfa_pool);
}


dfastate(nfa_states)
set nfa_states;		/* list of nfa states it could be */
{
	register int	n,p;

	/* use hash table to find where dfa state goes quickly */
	p = set_hash(nfa_states,hash_entries);
	while ( dfa_hash_table[p] != NIL_INDEX &&
		!set_equ(dfa_array[dfa_hash_table[p]].nfa_states,nfa_states)){
		if (++p >= hash_entries)
				p = 0;
		}
	if ((n = dfa_hash_table[p]) == NIL_INDEX){
		/* make new entry */
		n = NEW_DFA(nfa_states);
		dfa_hash_table[p] = n;
		}
	return n;
}


/* this reach assumes the closure has been done already on set */
set reach(b,a)
set b;
register int a;
{
	register int node;
	register nfa_node *p;
	register unsigned *start;
	set temp;

	start = &(b.setword[0]);
	temp = set_of(nil);
	set_remove_int(b,node,start);
	while (node != nil){
		p = &nfa(node);
		if (set_el(a,p->label))
			set_or_int(temp, p->trans[0]);
		set_remove_int(b,node,start);
		}
	return temp;
}

set closure(b)
set b;
{
	register int node;
	register unsigned *start;	/* keep track place in set */
	set undone;

	undone = b;
	start = &(undone.setword[0]);
	set_remove_int(undone,node,start);
	while (node != nil){
		/* find those epsilon transitions */
		if (nfa(node).trans[0] != NIL_INDEX &&
		    set_nil(nfa(node).label) &&
		    !set_el(nfa(node).trans[0],b)){
			set_or_int(undone,nfa(node).trans[0]);
			set_or_int(b,nfa(node).trans[0]);
			}
		if (nfa(node).trans[1] != NIL_INDEX &&
		    !set_el(nfa(node).trans[1],b)){
			set_or_int(undone,nfa(node).trans[1]);
			set_or_int(b,(nfa(node).trans[1]));
			}
		/* not sure where new nodes put in, so have to
		   start at beginning of set */
		start = &(undone.setword[0]);
		set_remove_int(undone,node,start);
		}
	return b;
}

int NEW_DFA(nfa_states)
set nfa_states;
{
	dfa_array[dfa_pool] = clear_dfa_node;
	dfa_array[dfa_pool].nfa_states = nfa_states;
	return dfa_pool++;
}

print_hash_table(begin,end)
int begin,end;
{
	register i = 0,t=0;

	for(i= begin; i<end; ++i){
		printf("[%d] = %d\t",i,dfa_hash_table[i]);
		if (++t > 4){
			t = 0;
			printf("\n");
			}
		}
}
